"""
Contains Zeitraum class
and corresponding marshmallow schema for de-/serialization
"""

from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Optional

from pydantic import validator

from bo4e.com.com import COM
from bo4e.enum.zeiteinheit import Zeiteinheit

# pylint: disable=too-few-public-methods


class Zeitraum(COM):
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher eine der drei Möglichkeiten angegeben sein:
    - Einheit und Dauer oder
    - Zeitraum: Startdatum bis Enddatum oder
    - Zeitraum: Startzeitpunkt (Datum und Uhrzeit) bis Endzeitpunkt (Datum und Uhrzeit)

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zeitraum.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zeitraum JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Zeitraum.json>`_

    """

    # optional attributes
    einheit: Optional[Zeiteinheit] = None
    dauer: Optional[Decimal] = None
    startdatum: Optional[datetime] = None
    enddatum: Optional[datetime] = None
    startzeitpunkt: Optional[datetime] = None
    endzeitpunkt: Optional[datetime] = None

    # pylint: disable=unused-argument, no-self-argument
    @validator("endzeitpunkt", always=True)
    def time_range_possibilities(cls, endzeitpunkt: Optional[datetime], values: Dict[str, Any]) -> Optional[datetime]:
        """
        An address is valid if it contains a postfach XOR (a strasse AND hausnummer).
        This functions checks for these conditions of a valid address.
        """

        if (
            values["einheit"]
            and values["dauer"]
            and not (values["startdatum"] or values["enddatum"] or values["startzeitpunkt"] or endzeitpunkt)
        ):
            return endzeitpunkt
        if (
            values["startdatum"]
            and values["enddatum"]
            and not (values["einheit"] or values["dauer"] or values["startzeitpunkt"] or endzeitpunkt)
        ):
            return endzeitpunkt
        if (
            values["startzeitpunkt"]
            and endzeitpunkt
            and not (values["einheit"] or values["dauer"] or values["startdatum"] or values["enddatum"])
        ):
            return endzeitpunkt

        raise ValueError(
            """
            Please choose from one of the three possibilities to specify the timerange:
            - einheit and dauer
            - startdatum and enddatum
            - startzeitpunkt and endzeitpunkt
            """
        )



"""
Contains Rechnungsposition class and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime

# pylint: disable=too-few-public-methods, too-many-instance-attributes
from typing import Any, Dict, Optional

from pydantic import validator

from bo4e.com.betrag import Betrag
from bo4e.com.com import COM
from bo4e.com.menge import Menge
from bo4e.com.preis import Preis
from bo4e.com.steuerbetrag import Steuerbetrag
from bo4e.enum.bdewartikelnummer import BDEWArtikelnummer
from bo4e.enum.zeiteinheit import Zeiteinheit
from bo4e.validators import check_bis_is_later_than_von


class Rechnungsposition(COM):
    """
    Über Rechnungspositionen werden Rechnungen strukturiert.
    In einem Rechnungsteil wird jeweils eine in sich geschlossene Leistung abgerechnet.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Rechnungsposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `Rechnungsposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Rechnungsposition.json>`_

    """

    # required attributes
    #: Fortlaufende Nummer für die Rechnungsposition
    positionsnummer: int

    lieferung_von: datetime  #: Start der Lieferung für die abgerechnete Leistung (inklusiv)
    lieferung_bis: datetime  #: Ende der Lieferung für die abgerechnete Leistung (exklusiv)
    _bis_check = validator("lieferung_bis", always=True, allow_reuse=True)(check_bis_is_later_than_von)

    #: Bezeichung für die abgerechnete Position
    positionstext: str

    #: Die abgerechnete Menge mit Einheit
    positions_menge: Menge
    #: Der Preis für eine Einheit der energetischen Menge
    einzelpreis: Preis

    teilsumme_netto: Betrag
    """
    Das Ergebnis der Multiplikation aus einzelpreis * positionsMenge * (Faktor aus zeitbezogeneMenge).
    Z.B. 12,60€ * 120 kW * 3/12 (für 3 Monate).
    """
    # the cross check in general doesn't work because Betrag and Preis use different enums to describe the currency
    # see https://github.com/Hochfrequenz/BO4E-python/issues/126

    #: Auf die Position entfallende Steuer, bestehend aus Steuersatz und Betrag
    teilsumme_steuer: Steuerbetrag

    # optional attributes
    #: Falls sich der Preis auf eine Zeit bezieht, steht hier die Einheit
    zeiteinheit: Optional[Zeiteinheit] = None

    #: Kennzeichnung der Rechnungsposition mit der Standard-Artikelnummer des BDEW
    artikelnummer: Optional[BDEWArtikelnummer] = None
    #: Marktlokation, die zu dieser Position gehört
    lokations_id: Optional[str] = None

    zeitbezogene_menge: Optional[Menge] = None
    """
    Eine auf die Zeiteinheit bezogene Untermenge.
    Z.B. bei einem Jahrespreis, 3 Monate oder 146 Tage.
    Basierend darauf wird der Preis aufgeteilt.
    """
    #: Nettobetrag für den Rabatt dieser Position
    teilrabatt_netto: Optional[Betrag] = None

    #: Standardisierte vom BDEW herausgegebene Liste, welche im Strommarkt die BDEW-Artikelnummer ablöst
    artikel_id: Optional[str] = None

    @staticmethod
    def _get_inclusive_start(values: Dict[str, Any]) -> datetime:
        """return the inclusive start (used in the validator)"""
        return values["lieferung_von"]

    # def _get_exclusive_end(self) -> datetime:
    #     """return the exclusive end (used in the validator)"""
    #     return self.lieferung_bis



"""
Contains Zaehler class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from decimal import Decimal

# pylint: disable=unused-argument
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.zaehlwerk import Zaehlwerk
from bo4e.enum.botyp import BoTyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.zaehlerauspraegung import Zaehlerauspraegung
from bo4e.enum.zaehlertyp import Zaehlertyp

# pylint: disable=too-many-instance-attributes, too-few-public-methods


class Zaehler(Geschaeftsobjekt):
    """
    Object containing information about a meter/"Zaehler".

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Zaehler.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zaehler JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Zaehler.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.ZAEHLER
    zaehlernummer: str  #: Nummerierung des Zählers,vergeben durch den Messstellenbetreiber
    sparte: Sparte  #: Strom oder Gas
    zaehlerauspraegung: Zaehlerauspraegung  #: Spezifikation die Richtung des Zählers betreffend
    zaehlertyp: Zaehlertyp  #: Typisierung des Zählers
    zaehlwerke: conlist(Zaehlwerk, min_items=1)  # type: ignore[valid-type]  #: Die Zählwerke des Zählers
    tarifart: Tarifart  #: Spezifikation bezüglich unterstützter Tarifarten

    # optional attributes
    zaehlerkonstante: Optional[Decimal] = None  #: Zählerkonstante auf dem Zähler
    eichung_bis: Optional[datetime] = None  #: Bis zu diesem Datum (exklusiv) ist der Zähler geeicht.
    letzte_eichung: Optional[datetime] = None  #: Zu diesem Datum fand die letzte Eichprüfung des Zählers statt.
    zaehlerhersteller: Optional[Geschaeftspartner] = None  #: Der Hersteller des Zählers



"""
Contains Messlokation class
and corresponding marshmallow schema for de-/serialization
"""
import re
from typing import Any, Dict, List, Optional

from iso3166 import countries

# Structure of a Messlokations-ID
# Ländercode nach DIN ISO 3166 (2 Stellen)
# Verteilnetzbetreiber (6 Stellen)
# Postleitzahl (5 Stellen)
# Zählpunktnummer (20 Stellen alphanumerisch)
# source: https://de.wikipedia.org/wiki/Z%C3%A4hlpunkt#Struktur_der_Z%C3%A4hlpunktbezeichnung
from pydantic import validator

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.zaehler import Zaehler
from bo4e.com.adresse import Adresse
from bo4e.com.dienstleistung import Dienstleistung
from bo4e.com.geokoordinaten import Geokoordinaten
from bo4e.com.hardware import Hardware
from bo4e.com.katasteradresse import Katasteradresse
from bo4e.enum.botyp import BoTyp
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte

_melo_id_pattern = re.compile(r"^[A-Z]{2}\d{6}\d{5}[A-Z\d]{20}$")


# pylint: disable=too-many-instance-attributes, too-few-public-methods


class Messlokation(Geschaeftsobjekt):
    """
    Object containing information about a Messlokation

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Messlokation.svg" type="image/svg+xml"></object>

    .. HINT::
        `Messlokation JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Messlokation.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.MESSLOKATION
    #: Die Messlokations-Identifikation; Das ist die frühere Zählpunktbezeichnung
    messlokations_id: str
    #: Sparte der Messlokation, z.B. Gas oder Strom
    sparte: Sparte

    # optional attributes
    #: Spannungsebene der Messung
    netzebene_messung: Optional[Netzebene] = None
    #: Die Nummer des Messgebietes in der ene't-Datenbank
    messgebietnr: Optional[str] = None
    #: Liste der Hardware, die zu dieser Messstelle gehört
    geraete: Optional[List[Hardware]] = None
    #: Liste der Messdienstleistungen, die zu dieser Messstelle gehört
    messdienstleistung: Optional[List[Dienstleistung]] = None  # todo: rename to plural
    #: Zähler, die zu dieser Messlokation gehören
    messlokationszaehler: Optional[List[Zaehler]] = None

    # only one of the following two optional codenr attributes can be set
    grundzustaendiger_msb_codenr: Optional[str] = None
    """
    Codenummer des grundzuständigen Messstellenbetreibers, der für diese Messlokation zuständig ist.
    (Dieser ist immer dann Messstellenbetreiber, wenn kein anderer MSB die Einrichtungen an der Messlokation betreibt.)
    """
    grundzustaendiger_msbim_codenr: Optional[str] = None
    """
    Codenummer des grundzuständigen Messstellenbetreibers für intelligente Messsysteme, der für diese Messlokation
    zuständig ist.
    (Dieser ist immer dann Messstellenbetreiber, wenn kein anderer MSB die Einrichtungen an der Messlokation betreibt.)
    """
    # only one of the following three optional address attributes can be set
    messadresse: Optional[Adresse] = None
    """
    Die Adresse, an der die Messeinrichtungen zu finden sind.
    (Nur angeben, wenn diese von der Adresse der Marktlokation abweicht.)
    """
    geoadresse: Optional[Geokoordinaten] = None
    """
    Alternativ zu einer postalischen Adresse kann hier ein Ort mittels Geokoordinaten angegeben werden
    (z.B. zur Identifikation von Sendemasten).
    """
    katasterinformation: Optional[Katasteradresse] = None
    """
    Alternativ zu einer postalischen Adresse und Geokoordinaten kann hier eine Ortsangabe mittels Gemarkung und
    Flurstück erfolgen.
    """

    # pylint: disable=unused-argument, no-self-argument
    @validator("messlokations_id", always=True)
    def _validate_messlokations_id(cls, messlokations_id: str) -> str:
        if not messlokations_id:
            raise ValueError("The messlokations_id must not be empty.")
        if not _melo_id_pattern.match(messlokations_id):
            raise ValueError(f"The messlokations_id '{messlokations_id}' does not match {_melo_id_pattern.pattern}")
        if not messlokations_id[0:2] in countries:
            raise ValueError(f"The country code '{messlokations_id[0:2]}' is not a valid country code")
        return messlokations_id

    # pylint: disable=no-self-argument
    @validator("katasterinformation", always=True)
    def validate_address_info(
        cls, katasterinformation: Optional[Katasteradresse], values: Dict[str, Any]
    ) -> Optional[Katasteradresse]:
        """Checks that if an address is given, that there is only one valid address given"""
        all_address_attributes = [
            values["messadresse"],
            values["geoadresse"],
            katasterinformation,
        ]
        amount_of_given_address_infos = len([i for i in all_address_attributes if i is not None])
        if amount_of_given_address_infos > 1:
            raise ValueError("More than one address information is given.")
        return katasterinformation

    # pylint: disable=no-self-argument
    @validator("grundzustaendiger_msbim_codenr", always=True)
    def validate_grundzustaendiger_x_codenr(
        cls, grundzustaendiger_msbim_codenr: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        """Checks that if a codenr is given, that there is only one valid codenr given."""
        all_grundzustaendiger_x_codenr_attributes = [
            values["grundzustaendiger_msb_codenr"],
            grundzustaendiger_msbim_codenr,
        ]
        amount_of_given_grundzustaendiger_x_codenr = len(
            [i for i in all_grundzustaendiger_x_codenr_attributes if i is not None]
        )
        if amount_of_given_grundzustaendiger_x_codenr > 1:
            raise ValueError("More than one codenr is given.")
        return grundzustaendiger_msbim_codenr



# pylint: disable=missing-module-docstring,line-too-long
from bo4e.enum.strenum import StrEnum


class Rechnungsstatus(StrEnum):
    """
    Abbildung verschiedener Zustände, die im Rahmen der Rechnungsbearbeitung durchlaufen werden.
    """

    UNGEPRUEFT = "UNGEPRUEFT"  #: Die Rechnung wurde erstellt bzw ist eingegangen, wurde aber noch nicht geprüft.
    GEPRUEFT_OK = "GEPRUEFT_OK"  #: Die Rechnung wurde geprüft und als korrekt befunden.
    GEPRUEFT_FEHLERHAFT = "GEPRUEFT_FEHLERHAFT"  #: Bei der Prüfung der Rechnung hat sich herausgestellt, dass die Rechnung Fehler aufweist.
    GEBUCHT = "GEBUCHT"  #: Die Buchhaltung hat die Rechnung aufgenommen und berücksichtigt.
    BEZAHLT = "BEZAHLT"  #: Die Rechnung wurde beglichen.



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Rufnummernart(StrEnum):
    """
    Abbildung verschiedener Rufnummerntypen.
    """

    RUF_ZENTRALE = "RUF_ZENTRALE"  #: Rufnummer Zentrale
    FAX_ZENTRALE = "FAX_ZENTRALE"  #: Faxnummer Zentrale
    SAMMELRUF = "SAMMELRUF"  #: Sammelrufnummer
    SAMMELFAX = "SAMMELFAX"  #: Sammelfaxnummer
    ABTEILUNGRUF = "ABTEILUNGRUF"  #: Rufnummer Abteilung
    ABTEILUNGFAX = "ABTEILUNGFAX"  #: Faxnummer Abteilung
    RUF_DURCHWAHL = "RUF_DURCHWAHL"  #: Rufnummer Durchwahl
    FAX_DURCHWAHL = "FAX_DURCHWAHL"  #: Faxnummer Durchwahl
    MOBIL_NUMMER = "MOBIL_NUMMER"  #: Nummer des mobilen Telefons



"""
Contains Energiemenge class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.verbrauch import Verbrauch
from bo4e.enum.botyp import BoTyp
from bo4e.enum.lokationstyp import Lokationstyp


class Energiemenge(Geschaeftsobjekt):
    """
    Abbildung von Mengen, die Lokationen zugeordnet sind

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Energiemenge.svg" type="image/svg+xml"></object>

    .. HINT::
        `Energiemenge JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Energiemenge.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.ENERGIEMENGE
    #: Eindeutige Nummer der Marktlokation bzw. der Messlokation, zu der die Energiemenge gehört
    lokations_id: str
    # todo: add validator such that only mess- or marktlokations IDs are accepted + cross check with lokationstyp
    #: Gibt an, ob es sich um eine Markt- oder Messlokation handelt
    lokationstyp: Lokationstyp

    #: Gibt den Verbrauch in einer Zeiteinheit an
    energieverbrauch: conlist(Verbrauch, min_items=1)  # type: ignore[valid-type]
    # there are no optional attributes



import pytest
from pydantic import ValidationError

from bo4e.bo.zeitreihe import Zeitreihe
from bo4e.enum.medium import Medium
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.messart import Messart
from bo4e.enum.messgroesse import Messgroesse
from bo4e.enum.wertermittlungsverfahren import Wertermittlungsverfahren
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_zeitreihenwert import example_zeitreihenwert


class TestZeitreihe:
    @pytest.mark.parametrize(
        "zeitreihe",
        [
            pytest.param(
                Zeitreihe(
                    bezeichnung="Foo",
                    beschreibung="Bar",
                    version="0.0.1",
                    messgroesse=Messgroesse.BLINDLEISTUNG,
                    messart=Messart.MAXIMALWERT,
                    medium=Medium.STROM,
                    einheit=Mengeneinheit.KVARH,
                    wertherkunft=Wertermittlungsverfahren.MESSUNG,
                    werte=[example_zeitreihenwert],
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, zeitreihe: Zeitreihe) -> None:
        assert_serialization_roundtrip(zeitreihe)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitreihe()  # type: ignore[call-arg]

        assert "6 validation errors" in str(excinfo.value)



"""
Contains Angebot class and corresponding marshmallow schema for de-/serialization
"""

from datetime import datetime

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist, constr

from bo4e.bo.ansprechpartner import Ansprechpartner
from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.angebotsvariante import Angebotsvariante
from bo4e.enum.botyp import BoTyp
from bo4e.enum.sparte import Sparte


class Angebot(Geschaeftsobjekt):
    """
    Mit diesem BO kann ein Versorgungsangebot zur Strom- oder Gasversorgung oder die Teilnahme an einer Ausschreibung
    übertragen werden. Es können verschiedene Varianten enthalten sein (z.B. ein- und mehrjährige Laufzeit).
    Innerhalb jeder Variante können Teile enthalten sein, die jeweils für eine oder mehrere Marktlokationen erstellt
    werden.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Angebot.svg" type="image/svg+xml"></object>

    .. HINT::
        `Angebot JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Angebot.json>`_


    """

    bo_typ: BoTyp = BoTyp.ANGEBOT
    # required attributes
    #: Eindeutige Nummer des Angebotes
    angebotsnummer: constr(strict=True, regex=r"^\d+$")  # type: ignore[valid-type]
    #: Erstellungsdatum des Angebots
    angebotsdatum: datetime
    #: Sparte, für die das Angebot abgegeben wird (Strom/Gas)
    sparte: Sparte
    #: Ersteller des Angebots
    angebotsgeber: Geschaeftspartner
    #: Empfänger des Angebots
    angebotsnehmer: Geschaeftspartner

    varianten: conlist(Angebotsvariante, min_items=1)  # type: ignore[valid-type]
    """ Eine oder mehrere Varianten des Angebots mit den Angebotsteilen;
    Ein Angebot besteht mindestens aus einer Variante."""

    # optional attributes
    anfragereferenz: Optional[str] = None
    """	Referenz auf eine Anfrage oder Ausschreibung;
    Kann dem Empfänger des Angebotes bei Zuordnung des Angebotes zur Anfrage bzw. Ausschreibung helfen."""
    #: Bis zu diesem Zeitpunkt (Tag/Uhrzeit) inklusive gilt das Angebot
    bindefrist: Optional[datetime] = None
    #: Person, die als Angebotsnehmer das Angebot angenommen hat
    unterzeichner_angebotsnehmer: Optional[Ansprechpartner] = None
    #: Person, die als Angebotsgeber das Angebots ausgestellt hat
    unterzeichner_angebotsgeber: Optional[Ansprechpartner] = None



from datetime import datetime, timezone

import pytest
from pydantic import ValidationError

from bo4e.com.unterschrift import Unterschrift


class TestUnterschrift:
    def test_unterschrift_only_required_attributes(self) -> None:
        """
        Test de-/serialisation of Unterschrift with minimal attributes.
        """
        unterschrift = Unterschrift(name="Foo")

        json_string = unterschrift.json(by_alias=True, ensure_ascii=False)

        assert "Foo" in json_string

        unterschrift_deserialized = Unterschrift.parse_raw(json_string)

        assert isinstance(unterschrift_deserialized.name, str)
        assert unterschrift_deserialized.name == "Foo"

    def test_unterschrift_required_and_optional_attributes(self) -> None:
        """
        Test de-/serialisation of Unterschrift with maximal attributes.
        """
        unterschrift = Unterschrift(name="Foo", ort="Grünwald", datum=datetime(2019, 6, 7, tzinfo=timezone.utc))

        json_string = unterschrift.json(by_alias=True, ensure_ascii=False)

        assert "Foo" in json_string
        assert "Grünwald" in json_string
        assert "2019-06-07T00:00:00+00:00" in json_string

        unterschrift_deserialized = Unterschrift.parse_raw(json_string)

        assert isinstance(unterschrift_deserialized.name, str)
        assert unterschrift_deserialized.name == "Foo"
        assert isinstance(unterschrift_deserialized.ort, str)
        assert unterschrift_deserialized.ort == "Grünwald"
        assert isinstance(unterschrift_deserialized.datum, datetime)
        assert unterschrift_deserialized.datum == datetime(2019, 6, 7, tzinfo=timezone.utc)

    def test_unterschrift_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Unterschrift()  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)



"""
Contains Zaehlwerk class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal

# pylint: disable=no-name-in-module
# pylint: disable=no-name-in-module
from pydantic import constr

from bo4e.com.com import COM
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.mengeneinheit import Mengeneinheit

# pylint: disable=too-few-public-methods


class Zaehlwerk(COM):
    """
    Mit dieser Komponente werden Zählwerke modelliert.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zaehlwerk.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zaehlwerk JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Zaehlwerk.json>`_

    """

    zaehlwerk_id: str  # Identifikation des Zählwerks (Registers) innerhalb des Zählers.
    # Oftmals eine laufende Nummer hinter der Zählernummer. Z.B. 47110815_1
    bezeichnung: str  # Zusätzliche Bezeichnung, z.B. Zählwerk_Wirkarbeit.
    richtung: Energierichtung  # Die Energierichtung, Einspeisung oder Ausspeisung.
    obis_kennzahl: constr(  # type: ignore[valid-type]
        strict=True,
        regex=r"(?:(1)-((?:[0-5]?[0-9])|(?:6[0-5])):((?:[1-8]|99))\.((?:6|8|9|29))\.([0-9]{1,2}))|"
        r"(?:(7)-((?:[0-5]?[0-9])|(?:6[0-5])):(.{1,2})\.(.{1,2})\.([0-9]{1,2}))",
    )  # Die OBIS-Kennzahl für das Zählwerk, die festlegt, welche auf die gemessene Größe mit dem Stand gemeldet wird.
    # Nur Zählwerkstände mit dieser OBIS-Kennzahl werden an diesem Zählwerk registriert.
    wandlerfaktor: Decimal  # Mit diesem Faktor wird eine Zählerstandsdifferenz multipliziert, um zum eigentlichen Verbrauch im Zeitraum
    # zu kommen.
    einheit: Mengeneinheit  # Die Einheit der gemessenen Größe, z.B. kWh



import pytest
from pydantic import ValidationError

from bo4e.bo.tarifkosten import Tarifkosten
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.tarifmerkmal import Tarifmerkmal
from bo4e.enum.tariftyp import Tariftyp
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_energiemix import example_energiemix
from tests.test_kosten import example_kosten
from tests.test_marktteilnehmer import example_marktteilnehmer
from tests.test_vertragskonditionen import example_vertragskonditionen
from tests.test_zeitraum import example_zeitraum


class TestTarifkosten:
    @pytest.mark.parametrize(
        "tarifkosten",
        [
            pytest.param(
                Tarifkosten(
                    bezeichnung="foo",
                    anbietername="der beste stromanbieter",
                    sparte=Sparte.STROM,
                    kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                    tarifart=Tarifart.MEHRTARIF,
                    tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                    tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                    website="https://foo.inv",
                    bemerkung="super billig aber auch super dreckig",
                    vertragskonditionen=example_vertragskonditionen,
                    zeitliche_gueltigkeit=example_zeitraum,
                    energiemix=example_energiemix,
                    anbieter=example_marktteilnehmer,
                    kosten=example_kosten,
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifkosten: Tarifkosten) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifkosten)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarifkosten()  # type: ignore[call-arg]

        assert "9 validation errors" in str(excinfo.value)



from typing import List, Optional

import pytest
from pydantic import ValidationError

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.externereferenz import ExterneReferenz
from bo4e.enum.botyp import BoTyp


class TestGeschaeftsobjekt:
    @pytest.mark.parametrize(
        "bo_typ, versionstruktur, externe_referenzen",
        [
            (
                BoTyp.ENERGIEMENGE,
                "2",
                [
                    ExterneReferenz(ex_ref_name="HOCHFREQUENZ_HFSAP_100", ex_ref_wert="12345"),
                    ExterneReferenz(ex_ref_name="Schufa-ID", ex_ref_wert="aksdlakoeuhn"),
                ],
            ),
            (
                BoTyp.ENERGIEMENGE,
                "2",
                [ExterneReferenz(ex_ref_name="HOCHFREQUENZ_HFSAP_100", ex_ref_wert="12345")],
            ),
            (BoTyp.ENERGIEMENGE, "2", []),
        ],
    )
    def test_serialisation(
        self, bo_typ: BoTyp, versionstruktur: str, externe_referenzen: Optional[List[ExterneReferenz]]
    ) -> None:
        go = Geschaeftsobjekt(
            bo_typ=bo_typ,
            versionstruktur=versionstruktur,
            externe_referenzen=externe_referenzen,
        )
        assert isinstance(go, Geschaeftsobjekt)

        go_json = go.json(by_alias=True, ensure_ascii=False)

        assert str(versionstruktur) in go_json

        go_deserialized = Geschaeftsobjekt.parse_raw(go_json)

        assert go_deserialized.bo_typ is bo_typ
        assert go_deserialized.versionstruktur == versionstruktur
        assert go_deserialized.externe_referenzen == externe_referenzen

    def test_initialization_with_minimal_attributs(self) -> None:
        go = Geschaeftsobjekt(bo_typ=BoTyp.ANSPRECHPARTNER)

        assert go.externe_referenzen == []
        assert go.versionstruktur == "2"

    def test_no_list_in_externen_referenzen(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Geschaeftsobjekt(
                bo_typ=BoTyp.ENERGIEMENGE,
                externe_referenzen=ExterneReferenz(ex_ref_name="Schufa-ID", ex_ref_wert="aksdlakoeuhn"),  # type: ignore[arg-type]
            )
        assert "1 validation error" in str(excinfo.value)
        assert "value is not a valid list" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Preistyp(StrEnum):
    """
    Aufschlüsselung der Preistypen in Tarifen.
    """

    GRUNDPREIS = "GRUNDPREIS"  #: Grundpreis
    ARBEITSPREIS_EINTARIF = "ARBEITSPREIS_EINTARIF"  #: Arbeitspreis_ET
    ARBEITSPREIS_HT = "ARBEITSPREIS_HT"  #: Arbeitspreis_HT
    ARBEITSPREIS_NT = "ARBEITSPREIS_NT"  #: Arbeitspreis_NT
    LEISTUNGSPREIS = "LEISTUNGSPREIS"  #: Leistungspreis
    MESSPREIS = "MESSPREIS"  #: Messpreis
    ENTGELT_ABLESUNG = "ENTGELT_ABLESUNG"  #: Entgelt fuer Ablesung
    ENTGELT_ABRECHNUNG = "ENTGELT_ABRECHNUNG"  #: Entgelt fuer Abrechnung
    ENTGELT_MSB = "ENTGELT_MSB"  #: Entgelt für MSB (Entgelt für Einbau, Betrieb und Wartung der Messtechnik)
    PROVISION = "PROVISION"  #: Provision



from datetime import datetime, timezone

import pytest
from pydantic import ValidationError

from bo4e.bo.regionaltarif import Regionaltarif
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.tarifmerkmal import Tarifmerkmal
from bo4e.enum.tariftyp import Tariftyp
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_energiemix import example_energiemix
from tests.test_marktteilnehmer import example_marktteilnehmer
from tests.test_regionalepreisgarantie import example_regionale_preisgarantie
from tests.test_regionaleraufabschlag import example_regionaler_auf_abschlag
from tests.test_regionaletarifpreisposition import example_regionale_tarifpreisposition
from tests.test_tarifberechnungsparameter import example_tarifberechnungsparameter
from tests.test_tarifeinschraenkung import example_tarifeinschraenkung
from tests.test_vertragskonditionen import example_vertragskonditionen
from tests.test_zeitraum import example_zeitraum


class TestRegionaltarif:
    @pytest.mark.parametrize(
        "regionaltarif",
        [
            pytest.param(
                Regionaltarif(
                    preisstand=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                    berechnungsparameter=example_tarifberechnungsparameter,
                    tarif_auf_abschlaege=[example_regionaler_auf_abschlag],
                    tarifpreise=[example_regionale_tarifpreisposition],
                    preisgarantien=[example_regionale_preisgarantie],
                    tarifeinschraenkung=example_tarifeinschraenkung,
                    # ^^ above are the attributes of Regionaltarif
                    # vv below is all copy pasted from Tarifinfo test
                    bezeichnung="foo",
                    anbietername="der beste stromanbieter",
                    sparte=Sparte.STROM,
                    kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                    tarifart=Tarifart.MEHRTARIF,
                    tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                    tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                    website="https://foo.inv",
                    bemerkung="super billig aber auch super dreckig",
                    vertragskonditionen=example_vertragskonditionen,
                    zeitliche_gueltigkeit=example_zeitraum,
                    energiemix=example_energiemix,
                    anbieter=example_marktteilnehmer,
                ),
                id="required and optional attributes",
            ),
            pytest.param(
                Regionaltarif(
                    preisstand=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                    berechnungsparameter=example_tarifberechnungsparameter,
                    tarifpreise=[example_regionale_tarifpreisposition],
                    # ^^ above are the attributes of Regionaltarif
                    # vv below is all copy pasted from Tarifinfo test
                    bezeichnung="foo",
                    anbietername="der beste stromanbieter",
                    sparte=Sparte.STROM,
                    kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                    tarifart=Tarifart.MEHRTARIF,
                    tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                    tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                    anbieter=example_marktteilnehmer,
                ),
                id="required attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, regionaltarif: Regionaltarif) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(regionaltarif)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Regionaltarif()  # type: ignore[call-arg]
        assert "11 validation error" in str(excinfo.value)  # 3 from regionaltarif + 8 from tarifinfo

    def test_failing_validation_list_length_at_least_one(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Regionaltarif(
                preisstand=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                berechnungsparameter=example_tarifberechnungsparameter,
                tarifpreise=[],
                bezeichnung="foo",
                anbietername="der beste stromanbieter",
                sparte=Sparte.STROM,
                kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                tarifart=Tarifart.MEHRTARIF,
                tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                anbieter=example_marktteilnehmer,
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



from datetime import datetime, timezone
from decimal import Decimal
from typing import Dict

import pytest
from pydantic import ValidationError

from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.zeiteinheit import Zeiteinheit

example_zeitraum = Zeitraum(
    einheit=Zeiteinheit.TAG,
    dauer=Decimal(5),
)
example_zeitraum_dict = {
    "dauer": Decimal("5"),
    "startdatum": None,
    "endzeitpunkt": None,
    "einheit": Zeiteinheit.TAG,
    "enddatum": None,
    "startzeitpunkt": None,
}


class TestZeitraum:
    def test_zeitraum_dauer(self) -> None:
        """
        Test de-/serialisation of Zeitraum (only has optional attributes) with option dauer and einheit.
        """
        zeitraum = Zeitraum(einheit=Zeiteinheit.TAG, dauer=Decimal(21))

        json_string = zeitraum.json(by_alias=True, ensure_ascii=False)

        assert "21" in json_string
        assert "TAG" in json_string

        zeitraum_deserialized = Zeitraum.parse_raw(json_string)

        assert isinstance(zeitraum_deserialized.einheit, Zeiteinheit)
        assert zeitraum_deserialized.einheit == Zeiteinheit.TAG
        assert isinstance(zeitraum_deserialized.dauer, Decimal)
        assert zeitraum_deserialized.dauer == Decimal(21)

    def test_zeitraum_daten(self) -> None:
        """
        Test de-/serialisation of Zeitraum (only has optional attributes) with option startdatum and enddatum.
        """
        zeitraum = Zeitraum(
            startdatum=datetime(2013, 5, 1, tzinfo=timezone.utc), enddatum=datetime(2022, 1, 28, tzinfo=timezone.utc)
        )

        json_string = zeitraum.json(by_alias=True, ensure_ascii=False)

        assert "2013-05-01T00:00:00+00:00" in json_string
        assert "2022-01-28T00:00:00+00:00" in json_string

        zeitraum_deserialized = Zeitraum.parse_raw(json_string)

        assert isinstance(zeitraum_deserialized.startdatum, datetime)
        assert zeitraum_deserialized.startdatum == datetime(2013, 5, 1, tzinfo=timezone.utc)
        assert isinstance(zeitraum_deserialized.enddatum, datetime)
        assert zeitraum_deserialized.enddatum == datetime(2022, 1, 28, tzinfo=timezone.utc)

    def test_zeitraum_zeitpunkte(self) -> None:
        """
        Test de-/serialisation of Zeitraum (only has optional attributes) with option startzeitpunkt and endzeitpunkt.
        """
        zeitraum = Zeitraum(
            startzeitpunkt=datetime(2011, 2, 5, 16, 43, tzinfo=timezone.utc),
            endzeitpunkt=datetime(2021, 7, 30, tzinfo=timezone.utc),
        )

        json_string = zeitraum.json(by_alias=True, ensure_ascii=False)

        assert "2011-02-05T16:43:00+00:00" in json_string
        assert "2021-07-30T00:00:00+00:00" in json_string

        zeitraum_deserialized = Zeitraum.parse_raw(json_string)

        assert isinstance(zeitraum_deserialized.startzeitpunkt, datetime)
        assert zeitraum_deserialized.startzeitpunkt == datetime(2011, 2, 5, 16, 43, tzinfo=timezone.utc)
        assert isinstance(zeitraum_deserialized.endzeitpunkt, datetime)
        assert zeitraum_deserialized.endzeitpunkt == datetime(2021, 7, 30, tzinfo=timezone.utc)

    @pytest.mark.parametrize(
        "arguments",
        [
            pytest.param({"startdatum": datetime(2013, 5, 1, tzinfo=timezone.utc)}),
            pytest.param(
                {
                    "startdatum": datetime(2013, 5, 1, tzinfo=timezone.utc),
                    "startzeitpunkt": datetime(2011, 2, 5, 16, 43, tzinfo=timezone.utc),
                    "endzeitpunkt": datetime(2021, 7, 30, tzinfo=timezone.utc),
                }
            ),
            pytest.param({}),
            pytest.param(
                {
                    "startdatum": datetime(2013, 5, 1, tzinfo=timezone.utc),
                    "endzeitpunkt": datetime(2021, 7, 30, tzinfo=timezone.utc),
                }
            ),
        ],
    )
    def test_validator_time_range_possibilities(self, arguments: Dict[str, datetime]) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitraum(**arguments)  # type: ignore[arg-type]
        assert "Please choose from one of the three possibilities to specify the timerange:" in str(excinfo.value)



"""
Contains validators for BO s and COM s classes.
"""
import re
from datetime import datetime
from typing import Any, Dict, Optional, Protocol

from bo4e.enum.aufabschlagstyp import AufAbschlagstyp

# pylint: disable=unused-argument
from bo4e.enum.waehrungseinheit import Waehrungseinheit


def einheit_only_for_abschlagstyp_absolut(cls, value: Waehrungseinheit, values: Dict[str, Any]) -> Waehrungseinheit:  # type: ignore[no-untyped-def]
    """
    Check that einheit is only there if abschlagstyp is absolut.
    Currently, (2021-12-15) only used in COM AufAbschlag.
    """
    if value and (not values["auf_abschlagstyp"] or (values["auf_abschlagstyp"] != AufAbschlagstyp.ABSOLUT)):
        raise ValueError("Only state einheit if auf_abschlagstyp is absolute.")
    return value


# pylint:disable=too-few-public-methods
class _VonBisType(Protocol):
    """
    a protocol for all classes that have an inclusive start and exclusive end
    """

    @staticmethod
    def _get_inclusive_start(values: Dict[str, Any]) -> Optional[datetime]:
        """
        should return the inclusive start of the timeslice
        """

    # def _get_exclusive_end(self) -> Optional[datetime]:
    #     """
    #     should return the exclusive end of the timeslice
    #     """


def check_bis_is_later_than_von(cls, value: datetime, values: Dict[str, Any]):  # type: ignore[no-untyped-def]
    """
    assert that 'bis' is later than 'von'
    """
    # we want access to private methods here because these helper methods should be "hidden"
    start = cls._get_inclusive_start(values)  # pylint: disable=protected-access
    end = value
    if start and end and not end >= start:
        raise ValueError(f"The end '{end}' has to be later than the start '{start}'")
    return value


# pylint:disable=line-too-long
#: a regular expression that should match all OBIS Kennziffern
OBIS_PATTERN = r"((1)-((?:[0-5]?[0-9])|(?:6[0-5])):((?:[1-8]|99))\.((?:6|8|9|29))\.([0-9]{1,2})|(7)-((?:[0-5]?[0-9])|(?:6[0-5])):(.{1,2})\.(.{1,2})\.([0-9]{1,2}))"
# TODO: Maybe create custom data type for OBIS. See https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types

_malo_id_pattern = re.compile(r"^[1-9]\d{10}$")


# pylint: disable=unused-argument
def validate_marktlokations_id(cls, marktlokations_id: str, values: Dict[str, Any]) -> str:  # type: ignore[no-untyped-def]
    """
    A validator for marktlokations IDs
    """
    if not marktlokations_id:
        raise ValueError("The marktlokations_id must not be empty.")
    if not _malo_id_pattern.match(marktlokations_id):
        raise ValueError(f"The Marktlokations-ID '{marktlokations_id}' does not match {_malo_id_pattern.pattern}")
    expected_checksum = _get_malo_id_checksum(marktlokations_id)
    actual_checksum = marktlokations_id[10:11]
    if expected_checksum != actual_checksum:
        # pylint: disable=line-too-long
        raise ValueError(
            f"The Marktlokations-ID '{marktlokations_id}' has checksum '{actual_checksum}' but '{expected_checksum}' was expected."
        )
    return marktlokations_id


def _get_malo_id_checksum(malo_id: str) -> str:
    """
    Get the checksum of a marktlokations id.
    a) Quersumme aller Ziffern in ungerader Position
    b) Quersumme aller Ziffern auf gerader Position multipliziert mit 2
    c) Summe von a) und b) d) Differenz von c) zum nächsten Vielfachen von 10 (ergibt sich hier 10, wird die
       Prüfziffer 0 genommen
    https://bdew-codes.de/Content/Files/MaLo/2017-04-28-BDEW-Anwendungshilfe-MaLo-ID_Version1.0_FINAL.PDF
    :param self:
    :return: the checksum as string
    """
    odd_checksum: int = 0
    even_checksum: int = 0
    # start counting at 1 to be consistent with the above description
    # of "even" and "odd" but stop at tenth digit.
    for i in range(1, 11):
        digit = malo_id[i - 1 : i]
        if i % 2 - 1 == 0:
            odd_checksum += int(digit)
        else:
            even_checksum += 2 * int(digit)
    result: int = (10 - ((even_checksum + odd_checksum) % 10)) % 10
    return str(result)



from typing import Any, Dict

import pytest

from bo4e.com.geraet import Geraet
from bo4e.com.geraeteeigenschaften import Geraeteeigenschaften
from bo4e.enum.geraetemerkmal import Geraetemerkmal
from bo4e.enum.geraetetyp import Geraetetyp
from tests.serialization_helper import assert_serialization_roundtrip


class TestGeraet:
    @pytest.mark.parametrize(
        "geraet, expected_json_dict",
        [
            pytest.param(Geraet(), {"geraetenummer": None, "geraeteeigenschaften": None}, id="Minimal attributes"),
            pytest.param(
                Geraet(
                    geraetenummer="0815",
                    geraeteeigenschaften=Geraeteeigenschaften(
                        geraetemerkmal=Geraetemerkmal.GAS_G1000,
                        geraetetyp=Geraetetyp.MULTIPLEXANLAGE,
                    ),
                ),
                {
                    "geraetenummer": "0815",
                    "geraeteeigenschaften": {"geraetemerkmal": "GAS_G1000", "geraetetyp": Geraetetyp.MULTIPLEXANLAGE},
                },
                id="Maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, geraet: Geraet, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of Geraet
        """
        assert_serialization_roundtrip(geraet, expected_json_dict)



"""
Contains Region class and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.regionskriterium import Regionskriterium
from bo4e.enum.botyp import BoTyp


class Region(Geschaeftsobjekt):
    """
    Modellierung einer Region als Menge von Kriterien, die eine Region beschreiben

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Region.svg" type="image/svg+xml"></object>

    .. HINT::
        `Region JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Region.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.REGION
    #: Bezeichnung der Region
    bezeichnung: str

    #: Positivliste der Kriterien zur Definition der Region
    positiv_liste: conlist(Regionskriterium, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Negativliste der Kriterien zur Definition der Region
    negativ_liste: Optional[List[Regionskriterium]] = None



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.standorteigenschaftenstrom import StandorteigenschaftenStrom
from tests.serialization_helper import assert_serialization_roundtrip

example_standorteigenschaften_strom = StandorteigenschaftenStrom(
    regelzone="Transnet BW",
    bilanzierungsgebiet_eic="11YW-ALBSTADT--5",
    regelzone_eic="10YDE-ENBW-----N",
)


class TestStandorteigenschaftenStrom:
    @pytest.mark.parametrize(
        "standorteigenschaften_strom, expected_json_dict",
        [
            pytest.param(
                example_standorteigenschaften_strom,
                {
                    "regelzone": "Transnet BW",
                    "bilanzierungsgebietEic": "11YW-ALBSTADT--5",
                    "regelzoneEic": "10YDE-ENBW-----N",
                },
            )
        ],
    )
    def test_serialization_roundtrip(
        self, standorteigenschaften_strom: StandorteigenschaftenStrom, expected_json_dict: Dict[str, Any]
    ) -> None:
        assert_serialization_roundtrip(standorteigenschaften_strom, expected_json_dict)

    def test_missing_required_attributes(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = StandorteigenschaftenStrom()  # type: ignore[call-arg]

        assert "3 validation errors" in str(excinfo.value)
